/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "about.h"
#include "kemboss.h"

#include <QSettings>
#include <QFileDialog>

kembossWindow::kembossWindow(QMainWindow *parent) : QMainWindow(parent)
{
   setWindowIcon(QIcon(":/new/prefix1/kemboss.png"));
   fileToolBar = nullptr;
   editToolBar = nullptr;
   viewToolBar = nullptr;
   graphToolBar = nullptr;
   runToolBar = nullptr;
   helpToolBar = nullptr;
   kpwOutputGraphicsView = nullptr;
}

void kembossWindow::createActions()
{
   viewFileToolbarAct = new QAction(tr("File"), this);
   viewFileToolbarAct->setStatusTip(tr("View file toolbar"));
   viewFileToolbarAct->setCheckable(true);
   viewFileToolbarAct->setChecked(true);
   connect(viewFileToolbarAct, SIGNAL(triggered()), this, SLOT(viewFileToolbar()));

   viewEditToolbarAct = new QAction(tr("Edit"), this);
   viewEditToolbarAct->setStatusTip(tr("View edit toolbar"));
   viewEditToolbarAct->setCheckable(true);
   viewEditToolbarAct->setChecked(true);
   connect(viewEditToolbarAct, SIGNAL(triggered()), this, SLOT(viewEditToolbar()));

   viewViewToolbarAct = new QAction(tr("View"), this);
   viewViewToolbarAct->setStatusTip(tr("View view toolbar"));
   viewViewToolbarAct->setCheckable(true);
   viewViewToolbarAct->setChecked(true);
   connect(viewViewToolbarAct, SIGNAL(triggered()), this, SLOT(viewViewToolbar()));

   viewGraphToolbarAct = new QAction(tr("Graph"), this);
   viewGraphToolbarAct->setStatusTip(tr("View graph toolbar"));
   viewGraphToolbarAct->setCheckable(true);
   viewGraphToolbarAct->setChecked(true);
   connect(viewGraphToolbarAct, SIGNAL(triggered()), this, SLOT(viewGraphToolbar()));

   viewRunToolbarAct = new QAction(tr("Run"), this);
   viewRunToolbarAct->setStatusTip(tr("View run toolbar"));
   viewRunToolbarAct->setCheckable(true);
   viewRunToolbarAct->setChecked(true);
   connect(viewRunToolbarAct, SIGNAL(triggered()), this, SLOT(viewRunToolbar()));

   viewHelpToolbarAct = new QAction(tr("Help"), this);
   viewHelpToolbarAct->setStatusTip(tr("View help toolbar"));
   viewHelpToolbarAct->setCheckable(true);
   viewHelpToolbarAct->setChecked(true);
   connect(viewHelpToolbarAct, SIGNAL(triggered()), this, SLOT(viewHelpToolbar()));

   reportBugAct = new QAction(QIcon(":/new/prefix1/mail.png"), tr("&Report GUI Bug"), this);
   reportBugAct->setStatusTip(tr("Report a GUI bug"));
   connect(reportBugAct, SIGNAL(triggered()), this, SLOT(reportBug()));

   aboutAct = new QAction(QIcon(":/new/prefix1/about.png"), tr("About kemboss"), this);
   aboutAct->setStatusTip(tr("About kemboss"));
   connect(aboutAct, SIGNAL(triggered()), this, SLOT(aboutKemboss()));
}

void kembossWindow::readSettings()
{
    QSettings *settings;
    if (qobject_cast<kemboss *>(kembossParent)->ksettings.portableApp()) {
        settings = new QSettings(qobject_cast<kemboss *>(kembossParent)->ksettings.kembossPortableConfigDir()+objectName(), QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", objectName());
    }
    QPoint pos = settings->value("pos").toPoint();
    QSize size = settings->value("size").toSize();
    if (!size.isEmpty()) resize(size);
    if (!pos.isNull()) move(pos);
    // Read toolbar status
    if (fileToolBar != nullptr) showFileToolbar(settings->value("show_file_toolbar", true).toBool());
    if (editToolBar != nullptr) showEditToolbar(settings->value("show_edit_toolbar", true).toBool());
    if (viewToolBar != nullptr) showViewToolbar(settings->value("show_view_toolbar", true).toBool());
    if (graphToolBar != nullptr) {
        if (kpwOutputGraphicsView != nullptr) showGraphToolbar(settings->value("show_graph_toolbar", true).toBool());
        else showGraphToolbar(settings->value("show_graph_toolbar", false).toBool());
    }
    if (runToolBar != nullptr) showRunToolbar(settings->value("show_run_toolbar", true).toBool());
    if (helpToolBar != nullptr) showHelpToolbar(settings->value("show_help_toolbar", true).toBool());
    seqFormatInput = settings->value("highlight_input").toInt();
    seqFormatOutput = settings->value("highlight_output").toInt();
}

void kembossWindow::writeSettings()
{
    QSettings *settings;
    if (qobject_cast<kemboss *>(kembossParent)->ksettings.portableApp()) {
        settings = new QSettings(qobject_cast<kemboss *>(kembossParent)->ksettings.kembossPortableConfigDir()+objectName(), QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", objectName());
    }
    settings->setValue("pos", pos());
    settings->setValue("size", size());
    // Write toolbar status
    if (fileToolBar != nullptr) settings->setValue("show_file_toolbar", fileToolBar->isVisible());
    if (editToolBar != nullptr) settings->setValue("show_edit_toolbar", editToolBar->isVisible());
    if (viewToolBar != nullptr) settings->setValue("show_view_toolbar", viewToolBar->isVisible());
    if (graphToolBar != nullptr) settings->setValue("show_graph_toolbar", graphToolBar->isVisible());
    if (runToolBar != nullptr) settings->setValue("show_run_toolbar", runToolBar->isVisible());
    if (helpToolBar != nullptr) settings->setValue("show_help_toolbar", helpToolBar->isVisible());
    if (!curFile.isEmpty()) settings->setValue("last_settings", curFile);
    settings->setValue("highlight_input", seqFormatInput);
    settings->setValue("highlight_output", seqFormatOutput);
}

void kembossWindow::open()
{
   if (maybeSave()) {
      QString fileName = QFileDialog::getOpenFileName(this, tr("Open settings"),qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),objectName()+tr(" files (*.")+objectName()+tr(");;All files (*)"));
      if (!fileName.isEmpty())
         if (loadSettings(fileName))
            setCurrentFile(fileName);
   }
}

bool kembossWindow::save()
{
   if (curFile.isEmpty()) {
      return saveAs();
   } else {
      return saveSettings(curFile);
   }
}

bool kembossWindow::saveAs()
{
   QString fileName = QFileDialog::getSaveFileName(this, tr("Save settings"),qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),objectName()+tr(" files (*.")+objectName()+tr(");;All files (*)"));
   if (fileName.isEmpty())
      return false;
   if (!fileName.endsWith("."+objectName()))
      fileName.append("."+objectName());
   return saveSettings(fileName);
}

void kembossWindow::createRecentFiles()
{
   for (int i = 0; i < MaxRecentFiles; ++i) {
      recentFileActs[i] = new QAction(this);
      recentFileActs[i]->setVisible(false);
      connect(recentFileActs[i], SIGNAL(triggered()), this, SLOT(openRecentFile()));
      recentSettingsMenu->addAction(recentFileActs[i]);
    }
    updateRecentFileActions();
}

void kembossWindow::updateRecentFileActions()
{
    QSettings *settings;
    if (qobject_cast<kemboss *>(kembossParent)->ksettings.portableApp()) {
        settings = new QSettings(qobject_cast<kemboss *>(kembossParent)->ksettings.kembossPortableConfigDir()+objectName(), QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", objectName());
    }
    QStringList files = settings->value("recentFileList").toStringList();
    int numRecentFiles = qMin(files.size(), (int)MaxRecentFiles);

    for (int i = 0; i < numRecentFiles; ++i) {
        QString text = tr("&%1 %2").arg(i + 1).arg(strippedName(files[i]));
        recentFileActs[i]->setText(text);
        recentFileActs[i]->setData(files[i]);
        recentFileActs[i]->setVisible(true);
    }
    for (int j = numRecentFiles; j < MaxRecentFiles; ++j)
        recentFileActs[j]->setVisible(false);
}

void kembossWindow::openRecentFile()
{
   QAction *action = qobject_cast<QAction *>(sender());
   if (action)
      if (maybeSave())
         if (loadSettings(action->data().toString()))
            setCurrentFile(action->data().toString());
}

void kembossWindow::assistant(const QString &page)
{
    if (qobject_cast<kemboss *>(kembossParent)->ksettings.useOnlineDoc()) {
        QDesktopServices::openUrl(qobject_cast<kemboss *>(kembossParent)->ksettings.onlineDocSource()+page+".html");
    } else {
        if (QFile::exists(qobject_cast<kemboss *>(kembossParent)->ksettings.embossHelp()+page+".html"))
            QDesktopServices::openUrl(QUrl::fromLocalFile(qobject_cast<kemboss *>(kembossParent)->ksettings.embossHelp()+page+".html"));
        else if (QFile::exists(qobject_cast<kemboss *>(kembossParent)->ksettings.embossHelp()+"../../html/embassy/phylipnew/"+page+".html"))
            QDesktopServices::openUrl(QUrl::fromLocalFile(qobject_cast<kemboss *>(kembossParent)->ksettings.embossHelp()+"../../html/embassy/phylipnew/"+page+".html"));
    }
}

void kembossWindow::aboutKemboss()
{
   aboutForm *about = new aboutForm(qobject_cast<kemboss *>(kembossParent), &qobject_cast<kemboss *>(kembossParent)->ksettings);
   about->show();
}
