/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "kserver.h"
#include "sql.h"
#include <QFile>
#include <QMutex>
#include <QMutexLocker>
#include <QPointer>
#include <QCoreApplication>
#include <QDebug>

KembossServer::KembossServer(QObject *parent)
    : QTcpServer(parent)
{
    cserverStatus = kemboss::StartingUp; // Starting up, please wait. The server status is set to 'Ready' by the main event loop

    threadID = 0;
    threadsRunning = 0;

    inputThread = new KInputThread(this);
    connect(inputThread, SIGNAL(listRunningThreads()), this, SLOT(listRunningThreads()), Qt::QueuedConnection);
    connect(inputThread, SIGNAL(killThread(const quint32)), this, SLOT(killThread(const quint32)), Qt::QueuedConnection);
    connect(inputThread, SIGNAL(shutdownServer()), this, SLOT(shutdownServer()), Qt::QueuedConnection);
    inputThread->start();
}

KembossServer::~KembossServer()
{
    QSqlDatabase db = QSqlDatabase::database("kemboss", false);
    db.close();
}

bool KembossServer::loadEmbossProgramList(const QString &filename)
{
    qDebug() << "Loading EMBOSS program list";
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        return false;
    }
    QTextStream in(&file);
    while (!in.atEnd()) {
        cembossPrograms.append(in.readLine().trimmed());
    }
    return true;
}

void KembossServer::setPrivateKeyFile(const QString &filename)
{
    qDebug() << "Setting private key file.";
    if (!QFile::exists(filename)) {
        msgStdOut("Private key file " + filename + " not found. Please check Your server configuration.");
        return;
    }
    cprivateKeyFile = filename;
}

void KembossServer::setCACertsPath(const QString &path)
{
    qDebug() << "Setting CA certs path.";
    if (!path.isEmpty() && path != "none") {
        QDir dir;
        if (!dir.exists(path)) {
            msgStdOut("CA certificates path " + path + " not found. Please check Your server configuration.");
            cCACertsPath = "";
            return;
        } else cCACertsPath = path;
    } else cCACertsPath = "";
}

void KembossServer::setLocalCertFile(const QString &filename)
{
    qDebug() << "Setting local cert file.";
    if (!QFile::exists(filename)) {
        msgStdOut("Local certificate file " + filename + " not found. Please check Your server configuration.");
        return;
    }
    clocalCertFile = filename;
}

bool KembossServer::setEmbossFileDir(const QString &dir)
{
    qDebug() << "Setting EMBOSS file dir.";
    cembossFileDir = dir;
    if (!cembossFileDir.endsWith("/")) cembossFileDir += "/";
    QDir directory = QDir(cembossFileDir);
    if (!directory.exists()) {
        if (!directory.mkpath(cembossFileDir)) {
            msgStdErr("Could not create directory for users input and output files. "
                      "Maybe the user that is running the kemboss server does not have write access");
            return false;
        }
    }
    return true;
}

void KembossServer::incomingConnection(qintptr socketDescriptor)
{
    qDebug() << "Incoming Connection...";
    msgStdOut("Incoming Connection...");

    // As soon as a client connects, pass its incoming socket id to a new server thread
    QPointer<QThread> thread = new QThread(this);
    QPointer<KembossThread> connection = new KembossThread(socketDescriptor, this);

    threadsRunning++;
    threadID++;
    connection->threadID = threadID;

    connection->moveToThread(thread);

    QObject::connect(thread, SIGNAL(started()), connection, SLOT(run()), Qt::QueuedConnection); // When the thread is started the function run() of the connection is executed
    QObject::connect(connection, SIGNAL(threadFinished(QThread*)), this, SLOT(threadFinished(QThread*)), Qt::QueuedConnection); // The connection tells the server that the communication with the client is finished. The server deletes the connection and the thread
    QObject::connect(connection, SIGNAL(msgStdOut(QString)), this, SLOT(msgStdOut(QString)), Qt::QueuedConnection); // Allows the thread to write to StdOut
    QObject::connect(connection, SIGNAL(msgStdErr(QString)), this, SLOT(msgStdErr(QString)), Qt::QueuedConnection); // Allows the thread to write to StdErr
    QObject::connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater())); // When the thread is finished it is deleted

    threadList.append(connection);

    thread->start(); // Starting the thread executes run() in the connection class
}

bool KembossServer::embossProgramExecutionAllowed(const QString &program)
{
    QMutex mutex;
    QMutexLocker locker(&mutex);
    return cembossPrograms.contains(program, Qt::CaseInsensitive);
}


void KembossServer::threadFinished(QThread* thread) // The connection tells the server that the communication with the client is finished. The server deletes the connection and the thread
{
    QMutex mutex;
    mutex.lock();

    qDebug() << "Connection " << thread->currentThreadId() << " finished";

    int i = 0;
    while (i < threadList.size()) {
        if (threadList.at(i) != nullptr && threadList.at(i)->thread() == thread) {
            QString num;
            num.setNum(threadList.at(i)->threadID);
            msgStdOut("Connection "+num+" finished");
            threadList.at(i)->deleteLater(); // Deletes the connection object within the thread
            threadList.removeAt(i);
            break;
        }
        i++;
    }

    thread->quit(); // Finishes the thread
    if (!thread->wait(60000)) // If the thread doesn't finish within one minute it is terminated
        thread->terminate();

    threadsRunning--;

    mutex.unlock();
}

void KembossServer::listRunningThreads()
{
    QMutex mutex;
    mutex.lock();

    QTextStream qstdout(stdout);

    qstdout << "Running EMBOSS programs:" << endl;
    QString num;
    for (int i = 0; i < threadList.size(); ++i) {
        if (threadList.at(i) != nullptr) {
           num.setNum(threadList.at(i)->threadID);
           qstdout << num.leftJustified(5, ' ', true) + " "
                     + threadList.at(i)->getUserID().leftJustified(5, ' ', true) + " "
                     + threadList.at(i)->getEmbossApp().leftJustified(9, ' ', true) + " "
                     + threadList.at(i)->getRuntime() << endl;
        }
    }
    qstdout << endl;
    qstdout << "Enter server command: " << flush;

    mutex.unlock();
}

void KembossServer::killThread(const quint32 thread)
{
    QMutex mutex;
    mutex.lock();

    int i = 0;
    while (i < threadList.count() && threadList.at(i) != nullptr && threadList.at(i)->threadID != thread) i++;
    if (i < threadList.count() && threadList.at(i) != nullptr)
        threadList.at(i)->killEmbossProcess();

    mutex.unlock();
}

void KembossServer::shutdownServer()
{
    cserverStatus = kemboss::ShuttingDown;
    msgStdOut("Shutting down kemboss server...");

    // Waiting for the EMBOSS worker threads to finish
    while (!threadList.isEmpty()) {
        qDebug() << "Waiting for EMBOSS threads to finish...";
        inputThread->wait(1000);
    }

    inputThread->exit();
    close();
    exit(0);
}

void KembossServer::msgStdOut(const QString &msg)
{
    kembossServer::msgStdOut(msg, clogfile);
}

void KembossServer::msgStdErr(const QString &msg)
{
    kembossServer::msgStdErr(msg, clogfile);
}
