/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#ifndef KEMBOSSTHREAD_H
#define KEMBOSSTHREAD_H

#include <QSslSocket>
#include <QTimer>
#include <QProcess>
#include <QDir>
#include <QPointer>
#include <QElapsedTimer>
#include "ksshared.h"
#include "../shared.h"

class KembossThread : public QObject
{
    Q_OBJECT

public:
    KembossThread(qintptr socketDescriptor, QObject *parentServer = nullptr);
    virtual ~KembossThread();
    quint32 threadID;
    void setThreadStatus(const kembossServer::threadStatus threadStatus) { cthreadStatus = threadStatus; }
    kembossServer::threadStatus threadStatus() { return cthreadStatus; }

    QString getEmbossApp() { return embossApp; }
    QString getUserID() { QString num; num.setNum(userID); return num; }
    QString getRuntime() { QString timeElapsed; timeElapsed.setNum(runTime.elapsed()); return timeElapsed; }

    void killEmbossProcess();

public slots:
    void run();

signals:
    void error(QTcpSocket::SocketError socketError);
    void msgStdOut(const QString &msg);
    // Write a message to StdOut
    void msgStdErr(const QString &msg);
    // Write a message to StdErr
    void threadFinished(QThread* thread);
    // Inform the server that the thread is finished

private slots:
    void encryptedConnection();
    void tcpError(QAbstractSocket::SocketError error);
    void sslErrors(const QList<QSslError> &errors);
    // Display error in TCP connection
    void proxyAuthenticationRequired (const QNetworkProxy &, QAuthenticator *);
    // kemboss doesn't yet support connection via a network proxy
    void peerVerifyError(const QSslError &error);
    void readData();
    void readProgramOutput();
    void embossProcessFinished(int exitCode);
    void embossProcessError(QProcess::ProcessError error);
    void connectionClosed();
    void timeout();
    void stateChanged(QAbstractSocket::SocketState socketState);

private:
    bool startServerEncryption();
    // First we have to check if the client wants an SSL encrypted connection
    quint32 readUser(const QByteArray &data);
    // Receive and check user name and password. If all users are accepted, a new user is added to the database
    void readNewPassword(const QByteArray &data);
    // Receive new user password
    void readEmbossConfiguration(const QString &data);
    // Receive EMBOSS app, configuration and the names of the input file(s)
    void readEmbossOutputFileList(const QString &data);
    // Receive the list of EMBOSS output file(s)
    void readEmbossInputFileID(const QString &data);
    // Receive an emboss input filename (including path) followed by the md5 sum of the file.
    // If the file is not in the server cache, [get] is send to the client to receive the file. If the file is already in the server cache [has] is send to the client.
    // The filename (excluding path) is stored in inputFilename, the complete filename (including path) in clientInputFile the md5sum in inputMD5sum
    QString inputFilename;
    QString clientInputFile;
    QString inputMD5sum;
    void readEmbossInputFile(const QByteArray &data);
    // Receives a gzipped emboss input file and saves it in the inputDir directory with the filename inputFilename.
    // Then checks if the md5sum of the transferred file is identical to the md5sum received from the client (stored in inputMD5sum).
    // Then stores the filename, userID and md5sum of the file in the SQL database

    qint64 writeData(QByteArray data);
    // Sends data to the kemboss client
    void checkServer();
    // Returns the embassy applications installed on the server to the client
    void writeServerStatus();
    // Returns the server status to the client
    void writeEmbossOutput();
    // Returns the messages, the EMBOSS application produces to the client
    void writeEmbossOutputFileID();
    // Send EMBOSS output file identifier (filename (without path) followed by the md5 sum of the file)
    bool writeEmbossOutputFile(const QString &filename);
    // Transfer the output files produced by the EMBOSS application to the client

    QString checkIfFileIsCached(const QString &md5sum);
    // Checks if an EMBOSS input file is already cached on the server and returns the filename including the path. If the file is not already on the server an empty string is returned
    void userAccessDenied(const bool usernameFound);
    // This function is called when login to the server has failed. If the username has been found in the database but the password is incorrect usernameFound is true. If the username has not been found usernameFound is false
    bool checkEmbossApp();
    // Checks if EMBOSS program is allowed to be executed and if binary exists
    void runEmbossApp();
    // Run EMBOSS app
    void fileTransferFinished();
    // After all EMBOSS output files have been transferred to the client, the connection is terminated
    void userCanceled();
    // This function is invoked when the user cancels the execution of the EMBOSS program
    void logEmbossRun(const kembossServer::serverReturnCode returnCode);
    // Save the emboss application, user, runtime etc. to the database
    void closeConnection();
    // Close the TCP connection
    void abortConnection();
    // Abort the TCP connection

    QPointer<QObject> server;
    kembossServer::threadStatus cthreadStatus;
    QPointer<QSslSocket> tcpSocket;
    qintptr socketDescriptor;

    qint64 blockSize;

    quint32 userID;
    quint16 clientVersion;
    quint16 clientOS;

    QString embossApp;
    QStringList programConfiguration;

    QPointer<QProcess> programProcess;
    // The process, used to start the external EMBOSS application
    QByteArray cprogramOutput;
    // Contains the output of the EMBOSS program on stdout
    QDir serverInputDir;
    // The EMBOSS input file directory, equivalent to /embossFileDir/userName/input_files/
    QDir serverOutputDir;
    // The EMBOSS output directory, equivalent to /embossFileDir/userName/datetime.embossApp/
    QStringList clientEmbossOutputFiles;
    // Contains the filename(s) of the EMBOSS output files including the path(s) to the file(s) on the client
    QStringList serverEmbossOutputFiles;
    // Contains the EMBOSS output files on the server that have to be transferred to the client

    QPointer<QTimer> timer;
    QElapsedTimer runTime;
    // Timer to measure program execution time
    QElapsedTimer lastDataTransfer;
    // Timer to measure last data transfer. If no EMBOSS process is running and no data has been transferred for one minute, the thread is termindated
    QElapsedTimer embossProcessRuntime;
    // If the embossProcessRuntime exceeds a user defined value (cmaxEmbossProcessTime), the thread is terminated
};

/*Client - server communication to use EMBOSS on a remote server:

- The kemboss client initiates a connection to the EMBOSS server.
- Upon successfull connection, the kemboss server sends the current server status to the client, using the command [sst]
  (Ready | Shutting down to restart, no new jobs accepted | Busy, no new jobs accepted | Error, no new jobs accepted). The server status is followed by a 'Welcome message'
- If the kemboss client receives the server status 'Ready', the client sends the kemboss client version number [cvn],
  followed by the username, user group, user email and user password hash, each separated by [spl].
  The server looks up the username and password in the database. Upon verification the server responds with user acknowledged [uak] or user denied [ude].
  The field 'last_login' in the database is updated (it could be used to find users that do not use the server anymore).
- Upon user verification, the kemboss client sends the EMBOSS program name [epn], the EMBOSS program configuration [epc] and the MD5 sums of the EMBOSS input files [md5] to the server.
  The server compares the transmitted MD5 sums with the MD5 sums of the files in the users working directory on the server (stored in the MySQL database).
  If the input files are already in the users working directory on the server, these files are used for EMBOSS,
  otherwise the EMBOSS input file(s) are transferred to the users directory on the server, the MD5 sum is created and saved in the MySQL database.
- The server executes the EMBOSS program, saves the EMBOSS output in the users directory on the server, sends the EMBOSS output to the client and closes the connection.
*/

/* Commands for client - server communication:
 * [cvn] client version number followed by the user name, user group, user email and user password hash
 * [pwu] new user password
 * [sep] client asks for EMBASSY packages installed on the server
 * [epn] EMBOSS program name followed by the EMBOSS program configuration
 * [fid] EMBOSS input filename (without path) followed by the md5 sum of the file
 * [eif] gzipped emboss input file
 * [rap] transfer of input files finished, run EMBOSS app
 * [eol] EMBOSS output file list
 * [ore] EMBOSS output on stdout / EMBOSS output file received, send next EMBOSS output file identifier
 * [oir] EMBOSS output file identifier received, send EMBOSS output file
 * [cpe] EMBOSS program execution canceled by user
 *
 * Commands for server - client communication:
 * [sst] sending server status
 * [scf] SSL connection failed
 * [iep] list of EMBASSY packages installed on the server
 * [uak] user acknowledged
 * [uwp] username was found in the database but the password was wrong
 * [ude] user denied
 * [npw] user acknowledged, update password
 * [pna] execution of program not allowed
 * [bnf] EMBOSS binary not found on the server
 * [eea] Execution of EMBOSS binary allowed. Send EMBOSS output file list
 * [eor] Output file list received, send EMBOSS input files
 * [has] EMBOSS input file is already stored in the file cache, no need to retrieve it
 * [get] send EMBOSS input file
 * [fin] transfer of EMBOSS input file succeeded
 * [eou] EMBOSS output on stdout
 * [fod] EMBOSS output file identifier (filename (without path) followed by the md5 sum of the file)
 * [eof] gzipped EMBOSS output file
 * [tof] transfer of output files finished / program executed successfully. Closing connection
*/

#endif // KEMBOSSTHREAD_H
